@@ -1,7 +1,7 @@
 /*
  * Copyright 2003-2005 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.91 2005/12/07 01:04:52 vapier Exp $
+ * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.97 2005/12/29 14:03:25 vapier Exp $
  *
  * Copyright 2003-2005 Ned Ludd        - <solar@gentoo.org>
  * Copyright 2004-2005 Mike Frysinger  - <vapier@gentoo.org>
@@ -12,17 +12,18 @@
 #include <sys/types.h>
 #include <libgen.h>
 #include <limits.h>
-#define __USE_GNU
 #include <string.h>
 #include <errno.h>
 #include <unistd.h>
 #include <sys/stat.h>
+#include <sys/mman.h>
+#include <fcntl.h>
 #include <dirent.h>
 #include <getopt.h>
 #include <assert.h>
 #include "paxinc.h"
 
-static const char *rcsid = "$Id: scanelf.c,v 1.91 2005/12/07 01:04:52 vapier Exp $";
+static const char *rcsid = "$Id: scanelf.c,v 1.97 2005/12/29 14:03:25 vapier Exp $";
 #define argv0 "scanelf"
 
 #define IS_MODIFIER(c) (c == '%' || c == '#')
@@ -38,7 +39,8 @@ static void usage(int status);
 static void parseargs(int argc, char *argv[]);
 static char *xstrdup(const char *s);
 static void *xmalloc(size_t size);
-static void xstrcat(char **dst, const char *src, size_t *curr_len);
+static void xstrncat(char **dst, const char *src, size_t *curr_len, size_t n);
+#define xstrcat(dst,src,curr_len) xstrncat(dst,src,curr_len,0)
 static inline void xchrcat(char **dst, const char append, size_t *curr_len);
 
 /* variables to control behavior */
@@ -66,8 +68,12 @@ static char *find_lib = NULL;
 static char *out_format = NULL;
 static char *search_path = NULL;
 static char gmatch = 0;
+static char printcache = 0;
 
 
+caddr_t ldcache = 0;
+size_t ldcache_size = 0;
+
 /* sub-funcs for scanelf_file() */
 static void scanelf_file_get_symtabs(elfobj *elf, void **sym, void **tab)
 {
@@ -479,11 +485,98 @@ static void scanelf_file_rpath(elfobj *elf, char *found_rpath, char **ret, size_
 	else if (!be_quiet)
 		xstrcat(ret, "  -  ", ret_len);
 }
+
+#define LDSO_CACHE_MAGIC "ld.so-"
+#define LDSO_CACHE_MAGIC_LEN (sizeof LDSO_CACHE_MAGIC -1)
+#define LDSO_CACHE_VER "1.7.0"
+#define LDSO_CACHE_VER_LEN (sizeof LDSO_CACHE_VER -1)
+#define FLAG_ANY            -1
+#define FLAG_TYPE_MASK      0x00ff
+#define FLAG_LIBC4          0x0000
+#define FLAG_ELF            0x0001
+#define FLAG_ELF_LIBC5      0x0002
+#define FLAG_ELF_LIBC6      0x0003
+#define FLAG_REQUIRED_MASK  0xff00
+#define FLAG_SPARC_LIB64    0x0100
+#define FLAG_IA64_LIB64     0x0200
+#define FLAG_X8664_LIB64    0x0300
+#define FLAG_S390_LIB64     0x0400
+#define FLAG_POWERPC_LIB64  0x0500
+#define FLAG_MIPS64_LIBN32  0x0600
+#define FLAG_MIPS64_LIBN64  0x0700
+
+static char *lookup_cache_lib(elfobj *, char *);
+static char *lookup_cache_lib(elfobj *elf, char *fname)
+{
+	int fd = 0;
+	char *strs;
+	static char buf[_POSIX_PATH_MAX] = "";
+	const char *cachefile = "/etc/ld.so.cache";
+	struct stat st;
+
+	typedef struct {
+		char magic[LDSO_CACHE_MAGIC_LEN];
+		char version[LDSO_CACHE_VER_LEN];
+		int nlibs;
+	} header_t;
+	header_t *header;
+
+	typedef struct {
+		int flags;
+		int sooffset;
+		int liboffset;
+	} libentry_t;
+	libentry_t *libent;
+
+	if (fname == NULL)
+		return NULL;
+
+	if (ldcache == 0) {
+		if (stat(cachefile, &st) || (fd = open(cachefile, O_RDONLY)) == -1)
+			return NULL;
+
+		/* cache these values so we only map/unmap the cache file once */
+		ldcache_size = st.st_size;
+		ldcache = mmap(0, ldcache_size, PROT_READ, MAP_SHARED, fd, 0);
+
+		close(fd);
+
+		if (ldcache == (caddr_t)-1)
+			return NULL;
+
+		if (memcmp(((header_t *) ldcache)->magic, LDSO_CACHE_MAGIC, LDSO_CACHE_MAGIC_LEN))
+			return NULL;
+		if (memcmp (((header_t *) ldcache)->version, LDSO_CACHE_VER, LDSO_CACHE_VER_LEN))
+			return NULL;
+	}
+
+	header = (header_t *) ldcache;
+	libent = (libentry_t *) (ldcache + sizeof(header_t));
+	strs = (char *) &libent[header->nlibs];
+
+	for (fd = 0; fd < header->nlibs; fd++) {
+		/* this should be more fine grained, but for now we assume that
+		 * diff arches will not be cached together.  and we ignore the
+		 * the different multilib mips cases. */
+		if (elf->elf_class == ELFCLASS64 && !(libent[fd].flags & FLAG_REQUIRED_MASK))
+			continue;
+		if (elf->elf_class == ELFCLASS32 && (libent[fd].flags & FLAG_REQUIRED_MASK))
+			continue;
+
+		if (strcmp(fname, strs + libent[fd].sooffset) != 0)
+			continue;
+		strncpy(buf, strs + libent[fd].liboffset, sizeof(buf));
+	}
+	return buf;
+}
+
+
 static const char *scanelf_file_needed_lib(elfobj *elf, char *found_needed, char *found_lib, int op, char **ret, size_t *ret_len)
 {
 	unsigned long i;
 	char *needed;
 	void *strtbl_void;
+	char *p;
 
 	if ((op==0 && !show_needed) || (op==1 && !find_lib)) return NULL;
 
@@ -513,6 +606,9 @@ static const char *scanelf_file_needed_lib(elfobj *elf, char *found_needed, char
 					if (op == 0) { \
 						if (!be_wewy_wewy_quiet) { \
 							if (*found_needed) xchrcat(ret, ',', ret_len); \
+							if (printcache) \
+								if ((p = lookup_cache_lib(elf, needed)) != NULL) \
+									needed = p; \
 							xstrcat(ret, needed, ret_len); \
 						} \
 						*found_needed = 1; \
@@ -646,9 +742,11 @@ static char *scanelf_file_soname(elfobj *elf, char *found_soname)
 static char *scanelf_file_sym(elfobj *elf, char *found_sym)
 {
 	unsigned long i;
+	char *ret;
 	void *symtab_void, *strtab_void;
 
 	if (!find_sym) return NULL;
+	ret = find_sym;
 
 	scanelf_file_get_symtabs(elf, &symtab_void, &strtab_void);
 
@@ -667,13 +765,26 @@ static char *scanelf_file_sym(elfobj *elf, char *found_sym)
 					printf("%s(%s) %5lX %15s %s\n", \
 					       ((*found_sym == 0) ? "\n\t" : "\t"), \
 					       elf->base_filename, \
-					       (long)sym->st_size, \
-					       (char *)get_elfstttype(sym->st_info), \
+					       (unsigned long)sym->st_size, \
+					       get_elfstttype(sym->st_info), \
 					       symname); \
 					*found_sym = 1; \
-				} else if ((strcmp(find_sym, symname) == 0) || \
-				           (strcmp(symname, versioned_symname) == 0)) \
-					(*found_sym)++; \
+				} else { \
+					char *this_sym, *next_sym; \
+					this_sym = find_sym; \
+					do { \
+						next_sym = strchr(this_sym, ','); \
+						if (next_sym == NULL) \
+							next_sym = this_sym + strlen(this_sym); \
+						if ((strncmp(this_sym, symname, (next_sym-this_sym)) == 0 && symname[next_sym-this_sym] == '\0') || \
+						    (strcmp(symname, versioned_symname) == 0)) { \
+							ret = this_sym; \
+							(*found_sym)++; \
+							goto break_out; \
+						} \
+						this_sym = next_sym + 1; \
+					} while (*next_sym != '\0'); \
+				} \
 			} \
 			++sym; \
 		} }
@@ -681,10 +792,11 @@ static char *scanelf_file_sym(elfobj *elf, char *found_sym)
 		FIND_SYM(64)
 	}
 
+break_out:
 	if (be_wewy_wewy_quiet) return NULL;
 
 	if (*find_sym != '*' && *found_sym)
-		return find_sym;
+		return ret;
 	if (be_quiet)
 		return NULL;
 	else
@@ -834,7 +946,13 @@ static void scanelf_file(const char *filename)
 		case 's': out = scanelf_file_sym(elf, &found_sym); break;
 		default: warnf("'%c' has no scan code?", out_format[i]);
 		}
-		if (out) xstrcat(&out_buffer, out, &out_len);
+		if (out) {
+			/* hack for comma delimited output like `scanelf -s sym1,sym2,sym3` */
+			if (out_format[i] == 's' && (tmp=strchr(out,',')) != NULL)
+				xstrncat(&out_buffer, out, &out_len, (tmp-out));
+			else
+				xstrcat(&out_buffer, out, &out_len);
+		}
 	}
 
 #define FOUND_SOMETHING() \
@@ -999,9 +1117,8 @@ static void scanelf_envpath()
 }
 
 
-
 /* usage / invocation handling functions */
-#define PARSE_FLAGS "plRmyxetrnibSs:gN:TaqvF:f:o:BhV"
+#define PARSE_FLAGS "plRmyxetrnLibSs:gN:TaqvF:f:o:BhV"
 #define a_argument required_argument
 static struct option const long_opts[] = {
 	{"path",      no_argument, NULL, 'p'},
@@ -1014,6 +1131,7 @@ static struct option const long_opts[] = {
 	{"textrel",   no_argument, NULL, 't'},
 	{"rpath",     no_argument, NULL, 'r'},
 	{"needed",    no_argument, NULL, 'n'},
+	{"ldcache",   no_argument, NULL, 'L'},
 	{"interp",    no_argument, NULL, 'i'},
 	{"bind",      no_argument, NULL, 'b'},
 	{"soname",    no_argument, NULL, 'S'},
@@ -1044,6 +1162,7 @@ static const char *opts_help[] = {
 	"Print TEXTREL information",
 	"Print RPATH information",
 	"Print NEEDED information",
+	"Resolve NEEDED information (use with -n)",
 	"Print INTERP information",
 	"Print BIND information",
 	"Print SONAME information",
@@ -1124,27 +1243,26 @@ static void parseargs(int argc, char *argv[])
 		}
 
 		case 's': {
-			size_t len;
-			if (find_sym) err("Don't specify -s twice");
-			find_sym = xstrdup(optarg);
-			len = strlen(find_sym) + 1;
-			versioned_symname = (char*)xmalloc(sizeof(char) * (len+1));
+			if (find_sym) warn("You prob don't want to specify -s twice");
+			find_sym = optarg;
+			versioned_symname = (char*)xmalloc(sizeof(char) * (strlen(find_sym)+1+1));
 			sprintf(versioned_symname, "%s@", find_sym);
 			break;
 		}
 		case 'N': {
-			if (find_lib) err("Don't specify -N twice");
-			find_lib = xstrdup(optarg);
+			if (find_lib) warn("You prob don't want to specify -N twice");
+			find_lib = optarg;
 			break;
 		}
 
 		case 'F': {
-			if (out_format) err("Don't specify -F twice");
-			out_format = xstrdup(optarg);
+			if (out_format) warn("You prob don't want to specify -F twice");
+			out_format = optarg;
 			break;
 		}
 
-		case 'g': gmatch = 1;
+		case 'g': gmatch = 1; /* break; any reason we dont breal; here ? */
+		case 'L': printcache = 1; break;
 		case 'y': scan_symlink = 0; break;
 		case 'B': show_banner = 0; break;
 		case 'l': scan_ldpath = 1; break;
@@ -1243,14 +1361,12 @@ static void parseargs(int argc, char *argv[])
 	}
 
 	/* clean up */
-	if (find_sym) {
-		free(find_sym);
-		free(versioned_symname);
-	}
-	if (find_lib) free(find_lib);
-	if (out_format) free(out_format);
+	if (versioned_symname) free(versioned_symname);
 	for (i = 0; ldpaths[i]; ++i)
 		free(ldpaths[i]);
+
+	if (ldcache != 0)
+		munmap(ldcache, ldcache_size);
 }
 
 
@@ -1262,15 +1378,13 @@ static char *xstrdup(const char *s)
 	if (!ret) err("Could not strdup(): %s", strerror(errno));
 	return ret;
 }
-
 static void *xmalloc(size_t size)
 {
 	void *ret = malloc(size);
 	if (!ret) err("Could not malloc() %li bytes", (unsigned long)size);
 	return ret;
 }
-
-static void xstrcat(char **dst, const char *src, size_t *curr_len)
+static void xstrncat(char **dst, const char *src, size_t *curr_len, size_t n)
 {
 	size_t new_len;
 
@@ -1279,12 +1393,14 @@ static void xstrcat(char **dst, const char *src, size_t *curr_len)
 		*curr_len = new_len + (*curr_len / 2);
 		*dst = realloc(*dst, *curr_len);
 		if (!*dst)
-			err("could not realloc %li bytes", (unsigned long)*curr_len);
+			err("could not realloc() %li bytes", (unsigned long)*curr_len);
 	}
 
-	strcat(*dst, src);
+	if (n)
+		strncat(*dst, src, n);
+	else
+		strcat(*dst, src);
 }
-
 static inline void xchrcat(char **dst, const char append, size_t *curr_len)
 {
 	static char my_app[2];
